PROJECTS = "projects"
TOTAL_COUNT = "totalCount"
PAGE_INFO = "pageInfo"
HAS_NEXT_PAGE = "hasNextPage"
DATA = "data"
EXTENSIONS = "extensions"
SEVERITY = "severity"
WARNING = "WARNING"
ERRORS = "errors"
NODES = "nodes"
PROJECT_NAME_WITH_IN = "projectNameWithin"
WORKBOOKS_CONNECTION = "workbooksConnection"
PROJECT_LUID = "projectLuid"
NAME = "name"
ID = "id"
PROJECT_NAME = "projectName"
SHEETS = "sheets"
DASHBOARDS = "dashboards"
EMBEDDED_DATA_SOURCES = "embeddedDatasources"
UPSTREAM_COLUMNS = "upstreamColumns"
TABLE = "table"
TYPE_NAME = "__typename"
CUSTOM_SQL_TABLE = "CustomSQLTable"
UPSTREAM_TABLES = "upstreamTables"
DATABASE_TABLES_CONNECTION = "databaseTablesConnection"
DATABASE_SERVERS_CONNECTION = "databaseServersConnection"
HOST_NAME = "hostName"
FIELDS = "fields"
UPSTREAM_DATA_SOURCES = "upstreamDatasources"
COLUMNS = "columns"
COLUMNS_CONNECTION = "columnsConnection"
DATABASE = "database"
SCHEMA = "schema"
FULL_NAME = "fullName"
CONNECTION_TYPE = "connectionType"
IS_EMBEDDED = "isEmbedded"
UPSTREAM_FIELDS = "upstreamFields"
DATA_SOURCE = "datasource"
DATA_SOURCE_FIELD = "DatasourceField"
COLUMN_FIELD = "ColumnField"
IDENTITY = "IDENTITY"
CALCULATED_FIELD = "CalculatedField"
FORMULA = "formula"
CUSTOM_SQL_TABLE_CONNECTION = "customSQLTablesConnection"
ID_WITH_IN = "idWithin"
DATA_SOURCES = "datasources"
WORKBOOK = "workbook"
DATASET = "dataset"
TABLES = "tables"
DESCRIPTION = "description"
SQL = "SQL"
QUERY = "query"
SUB_TYPES = "subTypes"
VIEW = "view"
CUSTOM_SQL = "Custom SQL"
REMOTE_TYPE = "remoteType"
UNKNOWN = "UNKNOWN"
PUBLISHED_DATA_SOURCE = "PublishedDatasource"
LUID = "luid"
EMBEDDED_DATA_SOURCE = "EmbeddedDatasource"
UPSTREAM_LINEAGE = "upstreamLineage"
OWNER = "owner"
USERNAME = "username"
HAS_EXTRACTS = "hasExtracts"
EXTRACT_LAST_REFRESH_TIME = "extractLastRefreshTime"
EXTRACT_LAST_INCREMENTAL_UPDATE_TIME = "extractLastIncrementalUpdateTime"
EXTRACT_LAST_UPDATE_TIME = "extractLastUpdateTime"
PUBLISHED_DATA_SOURCES_CONNECTION = "publishedDatasourcesConnection"
FIELDS_CONNECTION = "fieldsConnection"
DATA_SOURCE_FIELDS = "datasourceFields"
SHEETS_CONNECTION = "sheetsConnection"
CREATED_AT = "createdAt"
UPDATED_AT = "updatedAt"
PATH = "path"
CONTAINED_IN_DASHBOARDS = "containedInDashboards"
TAGS = "tags"
CHART = "chart"
DASHBOARD = "dashboard"
DASHBOARDS_CONNECTION = "dashboardsConnection"
EMBEDDED_DATA_SOURCES_CONNECTION = "embeddedDatasourcesConnection"
PROJECT = "Project"
SITE = "Site"
IS_UNSUPPORTED_CUSTOM_SQL = "isUnsupportedCustomSql"
SITE_PERMISSION = "sitePermission"
ROLE_SITE_ADMIN_EXPLORER = "SiteAdministratorExplorer"
ROLE_SITE_ADMIN_CREATOR = "SiteAdministratorCreator"
ROLE_SERVER_ADMIN = "ServerAdministrator"
